defmodule ExRedi.Options do
  @moduledoc """
  Handles parsing/formatting of user-supplied options.

  **For Internal Use**
  """
  @type opts :: ExRedi.opts()

  @type key :: atom

  @languages ~w(
    arabic chinese danish dutch english finnish french
    german hungarian italian norwegian portuguese
    romanian russian spanish swedish tamil turkish
  )

  @options [
    create: [
      # Disables storing term offsets for documents
      {:nooffsets, :boolean},
      # Disables highlighting support
      {:nohl, :boolean},
      # Disables storing field bits for each term.
      {:nofields, :boolean},
      # Disables saving the term frequencies in the index
      {:nofreqs, :boolean},
      # Set a custom stopword list for the index
      {:stopwords, :list}
    ],
    drop: [
      # Disables deleting the actual document hashes
      {:keepdocs, :boolean}
    ],
    add: [
      # Disables saving the actual document in the index
      {:nosave, :boolean},
      # Performs an UPSERT style insertion
      {:replace, :boolean},
      # Does not require all fields for reindexing
      {:partial, :boolean},
      # Sets a binary safe payload string that can be evaluated at query time
      {:payload, :binary},
      # Uses a stemmer for the language during indexing
      {:language, :binary}
    ],
    add_hash: [
      # Performs an UPSERT style insertion
      {:replace, :boolean},
      # Uses a stemmer for the language during indexing
      {:language, :binary}
    ],
    search: [
      # Returns only the document ids and not the content
      {:nocontent, :boolean},
      # Limits which fields from the document are returned
      {:return, :list},
      # Limits the number of results returned
      # {offset} {num}
      {:limit, :list},
      # Filters the results to ones appearing only in specific document fields
      # {num} {field}
      {:infields, :list},
      # Limits the results to a specified set of keys
      # {num} {field}
      {:inkeys, :list},
      # Allows a number of unmatched offsets between phrase terms
      {:slop, :binary},
      # Ensures query terms in the document are the same order as in the query
      {:inorder, :boolean},
      # Limits results to those having numeric values between min and max
      # numeric_field min max:
      {:filter, :list},
      # Filters the results to a given radius from lon and lat
      # {geo_field} {lon} {lat} {raius} m|km|mi|ft
      {:geofilter, :list},
      # Disables filtering stopwords from the query
      {:nostopwords, :boolean},
      # Returns the relative internal score of each document
      {:withscores, :boolean},
      # Returns the value of the sorting key
      {:withsortkeys, :boolean},
      # Disables stemming for query expansion
      {:verbatim, :boolean},
      # Uses a stemmer for the langauge during search
      {:language, :binary},
      # Sets a custom query expander
      {:expander, :binary},
      # Sets a custom scoring function
      {:scorer, :binary},
      # Adds a binary safe payload that is exposed to custom scoring functions
      {:payload, :binary},
      # Returns payloads saved with the suggestions (or nil)
      {:withpayloads, :boolean},
      # Sorts results by a specified field
      # {field} [ASC|DESC]
      {:sortby, :list}
    ],
    aggregate: [
      # Groups results based on one or more properties
      # {nargs} {property} REDUCE {func} {nargs} {arg} … [AS {name}]
      {:groupby, :list},
      # Sorts results using a list of properties
      # {nargs} {property} {ASC|DESC} [MAX {num}]
      {:sortby, :list},
      # Applies a 1-to-1 transformation on one or more properties
      # {expr} AS {name}
      {:apply, :list},
      # Limits the number of results returned
      # {offset} {num}
      {:limit, :list}
    ],
    del: [
      # Deletes the actual document if it is in the index
      {:dd, :boolean}
    ],
    sugadd: [
      # Increments the existing score, instead of replacing it
      {:incr, :boolean},
      # Saves an extra payload with the suggestion
      {:payload, :binary}
    ],
    sugget: [
      # Performs a fuzzy prefix search
      {:fuzzy, :boolean},
      # Limits the number of results returned
      {:max, :binary},
      # Returns the score of each suggestion
      {:withscores, :boolean},
      # Returns payloads saved with the suggestions (or nil)
      {:withpayloads, :boolean}
    ]
  ]

  @keys Keyword.keys(@options)

  @spec build(key, opts) :: [binary | atom | integer | float]
  def build(key, opts) when key in @keys do
    @options
    |> Keyword.fetch!(key)
    |> Enum.map(fn {key, type} ->
      opts
      |> Keyword.get(key)
      |> guard_opt({key, type})
    end)
    |> Enum.reject(&is_nil/1)
    |> List.flatten()
  end

  defp guard_opt(value, {:language, _}) when value in @languages do
    [:language, value]
  end

  defp guard_opt(true, {key, :boolean}) do
    key
  end

  defp guard_opt(value, {key, :binary}) when is_binary(value) do
    [key, value]
  end

  defp guard_opt(value, {key, :list}) when is_list(value) do
    [key | value]
  end

  defp guard_opt(_, _) do
    nil
  end
end
